// Included C libraries
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>

// Utility function for timing in milliseconds
double diffclock(clock_t clock1, clock_t clock2)
// return difference in time in milliseconds
{
 double diffticks=clock1-clock2;
 double diffms=(diffticks*1000)/CLOCKS_PER_SEC;
 return diffms;
}

void DiffX_CPU(float* U, float* Ux, int N, int alpha, float* stencils, int rank) 
//
// CPU routine to compute spatial derivative in one space dimension.
//
// By Allan P. Engsig-Karup.
{
   // index in output vector Ux to be computed
   int row;
   for (row=0; row<N; ++row)
   { 
     float value=0.0;
     // Compute dot-product between FDM stencil weights and input vector U
     int diff = 0; // diff is used for automatically taking one-sided difference near boundaries
     if (row<alpha)
        diff = alpha - row;
     else if (row>N-1-alpha)  // row  >   Nx-3 Nx-2 Nx-1
        diff = N-1-alpha-row;
 
     int tmp = (alpha-diff)*rank+alpha;
     int tmp2 = row + diff;

     int i;
     for (i = -alpha; i<alpha+1; ++i)
       {
          value += U[tmp2+i]*stencils[tmp+i];
       }

     // Store computed approximation
     Ux[row] = value; 
   }
}

// External routine for computing finite difference stencils
extern void fdcoeffF(int k, float xbar, float x[], float c[], int n); 

int main(int argc, char *argv[])
{
    printf("Approximation of first derivative in 1D \n using the finite difference method.");

    int Nx, alpha;
    if (argc>1)
       Nx = strtol(argv[1],NULL,10);
    else
       Nx = 10;
    printf("Number of points in x-direction, Nx = %d. \n",Nx);

    if (argc>2)
       alpha = strtol(argv[2],NULL,10);
    else
       alpha = 2;
    printf("Halfwidth of finite difference stencil, alpha = %d. \n",alpha);

    int i,j;
    int ITERATIONS = 100; // number of times timed calculation is done for averaging

    int k = 1;
    int rank = 2*alpha+1;
    unsigned int size;
    float xbar;
	float* stencils = (float*)malloc(rank*rank * sizeof(float));
	float* c		= (float*)malloc(rank * sizeof(float));
	float* x		= (float*)malloc(rank * sizeof(float));
	float* U		= (float*)malloc(Nx * sizeof(float));
	float* Ux		= (float*)malloc(Nx * sizeof(float));
	float* grid	= (float*)malloc(Nx * sizeof(float));


    // Setup mesh
    for (i=0; i<Nx; ++i)
	{
	   grid[i] = (float) (i)/ (float) (Nx-1);
	}
    for (i=0; i<rank; ++i)
	{
	   x[i] = grid[i];
	}

    // Setup finite difference weights table
    for (i=0; i<rank; ++i)
	{
	   xbar = grid[i];
	   fdcoeffF(k,xbar,x,c,rank);
	   for (j=0; j<rank; ++j)
	      stencils[i*rank+j] = c[j]; // vector storage according to row major order
	}

    // Initialize input array
    for (i=0; i<Nx; ++i)
        {
            U[i] = 0.5*grid[i]*grid[i]; // defines u'(x) = x
        }

    // Initialize timer
    clock_t begin, end;

    // BENCHMARK cpu version for speedup estimates
    begin=clock();
    for (i=0; i<ITERATIONS; ++i)
       DiffX_CPU(U, Ux, Nx, alpha, stencils, rank); 
    end=clock();
    double cputime = 0.0;
    cputime = diffclock(end,begin) / (double) (ITERATIONS);
    printf("CPU Execution time (averaged over %d computations): %.6lf ms. \n \n",ITERATIONS,cputime);

    // Qualitative assessment of correctness...
    // output a few of the calculated gradients...
    printf("A few gradients computed by the CPU version…\n");
    for (i=Nx-5; i<Nx; ++i)
       printf("x[%d]=%f,   Ux[%d]=%f   \n",i,grid[i],i,Ux[i]);
    printf("\n");

	free(stencils);
	free(c);
	free(x);
	free(U);
	free(Ux);
	free(grid);
}

